(******************************************************************************
 *                             Core-extended                                  *
 *                                                                            *
 * Copyright (C) 2008- Jane Street Holding, LLC                               *
 *    Contact: opensource@janestreet.com                                      *
 *    WWW: http://www.janestreet.com/ocaml                                    *
 *                                                                            *
 *                                                                            *
 * This library is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU Lesser General Public                 *
 * License as published by the Free Software Foundation; either               *
 * version 2 of the License, or (at your option) any later version.           *
 *                                                                            *
 * This library is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 * Lesser General Public License for more details.                            *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public           *
 * License along with this library; if not, write to the Free Software        *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 *                                                                            *
 ******************************************************************************)

let measurement_clock = Posix_clock.Monotonic
let now () = Posix_clock.gettime measurement_clock

module Run_cost = struct
  type t = {
    run_time : int;
    gc_time  : int
  }

  let empty = {
    run_time = 0;
    gc_time  = 0;
  }

  let fold arr f init = 
    Array.fold_left (fun r v -> 
      {
        run_time = f r.run_time v.run_time;
        gc_time  = f r.gc_time v.gc_time;
      }) init arr

  let mean arr = 
    let sum = fold arr (+) empty in
    {
      run_time = sum.run_time / Array.length arr;
      gc_time  = sum.gc_time / Array.length arr;
    }

  let min arr = fold arr min arr.(0)
  let max arr = fold arr max arr.(0)
end

let stabilize_gc () =
  let rec loop failsafe last_heap_live_words =
    if failsafe <= 0 then 
      failwith "unable to stabilize the number of live words in the major heap";
    Gc.compact ();
    let stat = Gc.stat () in
    if stat.Gc.live_words <> last_heap_live_words then loop (failsafe - 1) stat.Gc.live_words
  in
  loop 10 0

let full_major_cost () =
  let count = 1_000 in
  let s = now () in
  for i = 1 to count do
    Gc.full_major ();
  done;
  let e = now () in
  (e - s) / count

let find_run_size gettime_cost f init =
  let rec loop samples =
    let s = now () in
    for i = 1 to samples do
      ignore (f init);
    done;
    let e = now () in
    (* we need enough samples so that the gettime_cost is < 1% of the cost of the run
       and we also demand that the total run take at least .5 seconds *)
    if gettime_cost > ((e - s) / 100) || (e - s) < 500 * 1000 * 1000 then loop (samples * 2)
    else (samples, e - s)
  in
  loop 1

let run_once ~f ~init ~sample_size ~gettime_cost ~full_major_cost =
  Printf.printf ".%!";
  let run_s = now () in
  for i = 1 to sample_size do
    ignore (f init);
  done;
  let run_e = now () in
  Gc.full_major ();
  let gc_e = now () in
  {Run_cost.
    run_time = (run_e - run_s - gettime_cost) / sample_size;
    gc_time  = (gc_e - run_e - full_major_cost) / sample_size;
  }

let bench f init =
  (* calculate how long it takes us to get a time measurement for the current thread *)
  
  Printf.printf "calculating cost of timing measurement: %!";
  let gettime_cost = 
    Posix_clock.mean_gettime_cost ~measure:measurement_clock ~using:Posix_clock.Monotonic
  in 
  Printf.printf "%i ns\n%!" gettime_cost;
  Printf.printf "calculating minimal measurable interval: %!";
  let gettime_min_interval = Posix_clock.min_interval measurement_clock in
  Printf.printf "%i ns\n%!" gettime_min_interval;
  (* find the number of samples of f needed before gettime cost is < 1% of the total *)
  Printf.printf "determining number of runs per sample: %!";
  let sample_size, run_time = find_run_size gettime_min_interval f init in
  Printf.printf "%i\n%!" sample_size;
  let runs = Array.create 100 Run_cost.empty in
  Printf.printf "stabilizing GC: %!";
  stabilize_gc ();
  Printf.printf "done\n%!";
  Printf.printf "calculating the cost of a full major sweep: %!";
  let full_major_cost = full_major_cost () in
  Printf.printf "%i ns\n%!" full_major_cost;
  Printf.printf "running samples (estimated time %i sec)\n%!" 
    ((run_time * 100) / 1000 / 1000 / 1000);
  for i = 0 to 99 do
    runs.(i) <- run_once ~f ~init ~sample_size ~gettime_cost ~full_major_cost
  done;
  Printf.printf "\n%!";
  let mean = Run_cost.mean runs in
  let min  = Run_cost.min runs in
  let max  = Run_cost.max runs in
  Printf.printf "mean: %i ns, gc: %i ns\n%!" mean.Run_cost.run_time mean.Run_cost.gc_time;
  Printf.printf "min: %i ns,  gc: %i ns\n%!" min.Run_cost.run_time min.Run_cost.gc_time;
  Printf.printf "max: %i ns,  gc: %i ns\n%!" max.Run_cost.run_time max.Run_cost.gc_time
